/*
Copyright (C) 2003-2004 Douglas Thain and the University of Wisconsin
Copyright (C) 2005- The University of Notre Dame
This software is distributed under a BSD-style license.
See the file COPYING for details.

This module written by James Fitzgerald, B.S. 2006.
*/

#ifdef HAS_FUSE

#define _FILE_OFFSET_BITS 64
#define FUSE_USE_VERSION 22

#include <fuse.h>
#include <pthread.h>

#include "auth_all.h"
#include "chirp_global.h"
#include "chirp_protocol.h"
#include "debug.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <errno.h>
#include <sys/statfs.h>

static int chirp_fuse_block_size = 65536;
static int chirp_fuse_timeout = 60;

static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;

static void parsepath( const char *path, char *newpath, char *host )
{
	int slashcnt = 0, n=0, pathn=0, serven=0;
	if(!strcmp(path, "/")) // if we're in the root directory
	{
		strcpy(host, "/");
		strcpy(newpath, "/");
	}
	else
	{
		while(path[n] != '\0')
		{
			if(path[n] == '/')
			{
				slashcnt++;
			}

			if(slashcnt == 1 && path[n] != '/') // part of server
			{
				host[serven] = path[n];
				serven++;
			}
			else if(slashcnt > 1)
			{
				newpath[pathn] = path[n];
				pathn++;
			}

			n++;
		}
	}
	newpath[pathn]='\0';
	host[serven] = '\0';
	
	if(slashcnt == 0) // if there are no slashes in the path...i.e. we are lsing the root dir
	{
		strcpy(host, path);
		strcpy(newpath, "/");
	}

}

static void chirp_stat_to_fuse_stat( struct chirp_stat *c, struct stat *f )
{
	memset(f,0,sizeof(*f));
	f->st_dev = c->cst_dev;
	f->st_ino = c->cst_ino;
	f->st_mode = c->cst_mode;
	f->st_nlink = c->cst_nlink;
	f->st_uid = getuid();
	f->st_gid = getgid();
	f->st_rdev = c->cst_rdev;
	f->st_size = c->cst_size;
	f->st_blksize = chirp_fuse_block_size;
	f->st_blocks = 1+c->cst_size/chirp_fuse_block_size;
	f->st_atime = c->cst_atime;
	f->st_mtime = c->cst_mtime;
	f->st_ctime = c->cst_ctime;
}

static int chirp_fuse_getattr(const char *path, struct stat *info)
{
	INT64_T result;
	struct chirp_stat cinfo;
	char newpath[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];

	parsepath(path,newpath,host); 

	pthread_mutex_lock(&mutex);
	result = chirp_global_stat(host, newpath, &cinfo, time(0)+chirp_fuse_timeout );
	pthread_mutex_unlock(&mutex);

	if(result<0) return -errno;

	chirp_stat_to_fuse_stat(&cinfo,info);

	return 0;
}

static int chirp_fuse_readlink(const char *path, char *buf, size_t size)
{
	INT64_T result;
	char newpath[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];
	parsepath(path,newpath,host);

	pthread_mutex_lock(&mutex);
	result = chirp_global_readlink(host, newpath, buf, size, time(0)+chirp_fuse_timeout );
	pthread_mutex_unlock(&mutex);

	if(result<0) return -errno;
	buf[result] = 0;

	return 0;
}

static int chirp_fuse_readdir( const char *path, void *buf, fuse_fill_dir_t filler, off_t offset, struct fuse_file_info *fi )
{
 	char newpath[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];
	struct chirp_dirstat *dir, *d;

	parsepath(path,newpath,host);

	pthread_mutex_lock(&mutex);
	dir = chirp_global_getdirstat(host,newpath,time(0)+chirp_fuse_timeout);
	pthread_mutex_unlock(&mutex);

	if(!dir) return -errno;

	for(d=dir;d;d=d->next) {
		struct stat info;
		chirp_stat_to_fuse_stat(&d->info,&info);
		filler(buf,d->name,&info,0);
	}

	pthread_mutex_lock(&mutex);
	chirp_global_freedirstat(dir);
	pthread_mutex_unlock(&mutex);

	return 0;
}

static int chirp_fuse_mkdir(const char *path, mode_t mode)
{
	INT64_T result;
	char newpath[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];
	parsepath(path,newpath,host);

	pthread_mutex_lock(&mutex);
	result = chirp_global_mkdir(host, newpath, mode, time(0) + chirp_fuse_timeout);
	pthread_mutex_unlock(&mutex);

	if(result<0) return -errno;
	return 0;
}

static int chirp_fuse_unlink(const char *path)
{
	INT64_T result;
	char newpath[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];
	parsepath(path,newpath,host);

	pthread_mutex_lock(&mutex);
	result = chirp_global_unlink(host, newpath, time(0) + chirp_fuse_timeout );
	pthread_mutex_unlock(&mutex);

	if(result<0) return -errno;
	return 0;
}

static int chirp_fuse_rmdir(const char *path)
{
	INT64_T result;
	char newpath[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];
	parsepath(path,newpath,host); 

	pthread_mutex_lock(&mutex);
	result = chirp_global_rmdir(host, newpath, time(0) + chirp_fuse_timeout );
	pthread_mutex_unlock(&mutex);

	if(result<0) return -errno;
	return 0;
}

static int chirp_fuse_symlink(const char *from, const char *to)
{
	INT64_T result;
	char frompath[CHIRP_PATH_MAX], topath[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];
	parsepath(from,frompath,host); 
	parsepath(to,topath,host);

	pthread_mutex_lock(&mutex);
	result = chirp_global_symlink( host, frompath, topath, time(0)+ chirp_fuse_timeout );
	pthread_mutex_unlock(&mutex);

	if(result<0) return -errno;
	return 0;
}

static int chirp_fuse_rename(const char *from, const char *to)
{
	INT64_T result;
	char frompath[CHIRP_PATH_MAX], topath[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];
	parsepath(from,frompath,host); 
	parsepath(to,topath,host);

	pthread_mutex_lock(&mutex);
	result = chirp_global_rename( host, frompath, topath, time(0)+ chirp_fuse_timeout );
	pthread_mutex_unlock(&mutex);

	if(result<0) return -errno;
	return 0;
}

static int chirp_fuse_link(const char *from, const char *to)
{
	INT64_T result;
	char frompath[CHIRP_PATH_MAX], topath[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];
	parsepath(from,frompath,host); 
	parsepath(to,topath,host);

	pthread_mutex_lock(&mutex);
	result = chirp_global_link( host, frompath, topath, time(0)+ chirp_fuse_timeout );
	pthread_mutex_unlock(&mutex);

	if(result<0) return -errno;
	return 0;
}

static int chirp_fuse_chmod(const char *path, mode_t mode)
{
	INT64_T result;
	char newpath[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];
	parsepath(path,newpath,host);

	pthread_mutex_lock(&mutex);
	result = chirp_global_chmod( host, newpath, mode, time(0)+chirp_fuse_timeout );
	pthread_mutex_unlock(&mutex);

	if(result<0) return -errno;

	return 0;
}

static int chirp_fuse_chown(const char *path, uid_t uid, gid_t gid)
{
	INT64_T result;
	char newpath[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];
	parsepath(path,newpath,host);

	pthread_mutex_lock(&mutex);
	result = chirp_global_chown( host, newpath, uid, gid, time(0)+chirp_fuse_timeout );
	pthread_mutex_unlock(&mutex);

	if(result<0) return -errno;

	return 0;
}

static int chirp_fuse_truncate(const char *path, off_t size)
{
	INT64_T result;
	char newpath[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];
	parsepath(path,newpath,host);

	pthread_mutex_lock(&mutex);
	result = chirp_global_truncate( host, newpath, size, time(0)+chirp_fuse_timeout );
	pthread_mutex_unlock(&mutex);

	if(result<0) return -errno;
	return 0;
}

static int chirp_fuse_utime(const char *path, struct utimbuf *buf)
{
	INT64_T result;
	char newpath[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];
	parsepath(path,newpath,host);
	time_t actime, modtime;

	pthread_mutex_lock(&mutex);
	result = chirp_global_utime( host, newpath, actime, modtime, time(0)+chirp_fuse_timeout );
	pthread_mutex_unlock(&mutex);

	if(result<0) return -errno;

	buf->actime = actime;
	buf->modtime = modtime;
	
	return 0;
}


static int chirp_fuse_open( const char *path, struct fuse_file_info *fi )
{
	struct chirp_file *file;
	int mode = 0;

	char newpath[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];

	parsepath(path,newpath,host); 

	pthread_mutex_lock(&mutex);
	file =  chirp_global_open( host, newpath, fi->flags, mode, time(0)+chirp_fuse_timeout);
	pthread_mutex_unlock(&mutex);

	if(!file) return -errno;

	fi->fh = (int)file;

	return 0;

}

static int chirp_fuse_release( const char *path, struct fuse_file_info *fi )
{
	struct chirp_file *file = (struct chirp_file *)fi->fh;

	pthread_mutex_lock(&mutex);
	chirp_global_close(file,time(0)+chirp_fuse_timeout);
	pthread_mutex_unlock(&mutex);

	fi->fh = 0;

	return 0;
}

static int chirp_fuse_read(const char *path, char *buf, size_t size, off_t offset, struct fuse_file_info *fi)
{
	struct chirp_file *file = (struct chirp_file *)fi->fh;
	INT64_T result;

	if(!file) return -EBADF;

	pthread_mutex_lock(&mutex);
	result = chirp_global_pread( file, buf, size, offset, time(0)+chirp_fuse_timeout );
	pthread_mutex_unlock(&mutex);

	if(result<0) return -errno;

	return result;
}

static int chirp_fuse_write(const char *path, const char *buf, size_t size, off_t offset, struct fuse_file_info *fi)
{
	struct chirp_file *file = (struct chirp_file *)fi->fh;
	INT64_T result;

	if(!file) return -EBADF;

	pthread_mutex_lock(&mutex);
	result = chirp_global_pwrite( file, buf, size, offset, time(0)+chirp_fuse_timeout );
	pthread_mutex_unlock(&mutex);

	if(result<0) return -errno;

	return result;
}

static int chirp_fuse_mknod(const char *path, mode_t mode, dev_t rdev)
{
	struct chirp_file *file;
	char newpath[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];

	parsepath(path,newpath,host);

	pthread_mutex_lock(&mutex);
	file =  chirp_global_open( host, newpath, O_CREAT, mode, time(0)+chirp_fuse_timeout);
	pthread_mutex_unlock(&mutex);

	if(!file) return -errno;

	pthread_mutex_lock(&mutex);
	chirp_global_close(file,time(0)+chirp_fuse_timeout);
	pthread_mutex_unlock(&mutex);

	return 0;
}


static int chirp_fuse_statfs(const char *path, struct statfs *info)
{
	INT64_T result;
	char newpath[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];
	parsepath(path,newpath,host);
	struct chirp_statfs cinfo;

	pthread_mutex_lock(&mutex);
	result = chirp_global_statfs( host, newpath, &cinfo, time(0)+chirp_fuse_timeout );
	pthread_mutex_unlock(&mutex);

	if(result<0) return -errno;

	memset(info,0,sizeof(*info));

	info->f_type = cinfo.f_type;
	info->f_bsize = cinfo.f_bsize;
	info->f_blocks = cinfo.f_blocks;
	info->f_bfree = cinfo.f_bfree;
	info->f_bavail = cinfo.f_bavail;
	info->f_files = cinfo.f_files;
	info->f_ffree = cinfo.f_ffree;

	return 0;
}

static struct fuse_operations chirp_fuse_operations = {
	.getattr	= chirp_fuse_getattr,
	.readlink	= chirp_fuse_readlink,
	.readdir	= chirp_fuse_readdir,
	.mknod		= chirp_fuse_mknod,
	.mkdir		= chirp_fuse_mkdir,
	.symlink	= chirp_fuse_symlink,
	.unlink		= chirp_fuse_unlink,
	.rmdir		= chirp_fuse_rmdir,
	.rename		= chirp_fuse_rename,
	.link		= chirp_fuse_link,
	.chmod		= chirp_fuse_chmod,
	.chown		= chirp_fuse_chown,
	.truncate	= chirp_fuse_truncate,
	.utime		= chirp_fuse_utime,
	.open		= chirp_fuse_open,
	.read		= chirp_fuse_read,
	.write		= chirp_fuse_write,
	.statfs		= chirp_fuse_statfs,
	.release	= chirp_fuse_release
};

int main(int argc, char *argv[])
{
	char *flags = getenv("CHIRP_FUSE_DEBUG_FLAGS");
	char *bs = getenv("CHIRP_FUSE_BLOCK_SIZE");
	char *auth = getenv("CHIRP_FUSE_AUTH_METHOD");
	char *tout = getenv("CHIRP_FUSE_TIMEOUT");

	debug_config(argv[0]);
	debug_config_file("/dev/stdout");

	if(flags) debug_flags_set(flags);
	if(bs)    chirp_fuse_block_size = atoi(bs);
	if(tout)  chirp_fuse_timeout = atoi(tout);
	if(auth) {
		auth_register_byname(auth);
	} else {
		auth_register_all();
	}

	return fuse_main(argc, argv, &chirp_fuse_operations);
}

#else

#include <stdio.h>

int main( int argc, char *argv[] )
{
	printf("%s: sorry, fuse support was not built in\n",argv[0]);
	return 1;
}

#endif
