/*
Copyright (C) 2003-2004 Douglas Thain and the University of Wisconsin
Copyright (C) 2005- The University of Notre Dame
This software is distributed under a BSD-style license.
See the file COPYING for details.
*/

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <dirent.h>
#include <termios.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "chirp_client.h"
#include "chirp_protocol.h"
#include "chirp_acl.h"
#include "chirp_group.h"

#include "timestamp.h"
#include "debug.h"
#include "auth_all.h"
#include "stringtools.h"
#include "xmalloc.h"
#include "list.h"
#include "domain_name_cache.h"
#include "md5.h"

#ifdef HAS_LIBREADLINE
#include "readline/readline.h"
#include "readline/history.h"
#endif

#if CCTOOLS_OPSYS_CYGWIN || CCTOOLS_OPSYS_DARWIN
#define fopen64 fopen
#define open64 open
#define lseek64 lseek
#define stat64 stat
#define fstat64 fstat
#define lstat64 lstat
#define fseeko64 fseeko
#endif

#define BUFFER_SIZE 32768

static int do_open( 	int argc, char **argv );
static int do_close( 	int argc, char **argv );
static int do_get( 	int argc, char **argv );
static int do_put( 	int argc, char **argv );
static int do_cat(      int argc, char **argv );
static int do_cd( 	int argc, char **argv );
static int do_lcd( 	int argc, char **argv );
static int do_pwd( 	int argc, char **argv );
static int do_lpwd( 	int argc, char **argv );
static int do_getacl( 	int argc, char **argv );
static int do_setacl( 	int argc, char **argv );
static int do_ls( 	int argc, char **argv );
static int do_mv( 	int argc, char **argv );
static int do_rm( 	int argc, char **argv );
static int do_mkdir( 	int argc, char **argv );
static int do_rmdir( 	int argc, char **argv );
static int do_stat( 	int argc, char **argv );
static int do_statfs( 	int argc, char **argv );
static int do_timeout( 	int argc, char **argv );
static int do_debug( 	int argc, char **argv );
static int do_help( 	int argc, char **argv );
static int do_quit( 	int argc, char **argv );
static int do_whoami( 	 int argc, char **argv );
static int do_whoareyou( int argc, char **argv );
static int do_exec( 	 int argc, char **argv );
static int do_run( 	 int argc, char **argv );
static int do_md5( 	 int argc, char **argv );
static int do_thirdput(  int argc, char **argv );
static int do_mkalloc(   int argc, char **argv );
static int do_lsalloc(   int argc, char **argv );
static int do_creategroup( int argc, char **argv );
static int do_adduser(     int argc, char **argv);
static int do_removeuser(  int argc, char **argv);
static int do_membership(  int argc, char** argv);
static int do_set_policy( int argc, char** argv);
static int do_put_recursive( const char *local_file, const char *remote_file );
static int do_get_recursive( const char *remote_file, const char *local_file );
static int process_command( int argc, char **argv );

static struct chirp_client *client=0;
static int timeout=3600;
static time_t stoptime=0;
static char current_host[CHIRP_PATH_MAX];
static char current_local_dir[CHIRP_PATH_MAX];
static char current_remote_dir[CHIRP_PATH_MAX];

static INT64_T xfer_items = 0;
static INT64_T xfer_bytes = 0;

static int long_information = 0;

struct command {
	char *name;
	int must_be_open;
	int minargs, maxargs;
	char *help;
	int (*handler) ( int argc, char **argv);
};

static struct command list[] =
{
{"open",    0, 1, 1, "<host>",                   do_open},
{"close",   1, 0, 0, "",                         do_close},
{"cat",     1, 1, 100, "<file> [file2] [file3] ...", do_cat},
{"cd",      1, 1, 1, "<remotedir>",              do_cd},
{"lcd",     0, 1, 1, "<localdir>",               do_lcd},
{"pwd",     1, 0, 0, "",                         do_pwd},
{"lpwd",    0, 0, 0, "",                         do_lpwd},
{"get",	    1, 1, 2, "<remotefile> [localfile]", do_get},
{"put",	    1, 1, 2, "<localfile> [remotefile]", do_put},
{"getacl",  1, 0, 1, "[remotepath]",             do_getacl},
{"listacl", 1, 0, 1, "[remotepath]",             do_getacl},
{"setacl",  1, 3, 3, "<remotepath> <user> <rwldva>",do_setacl},
{"ls",      1, 0, 2, "[-la] [remotepath]",       do_ls},
{"mv",      1, 2, 2, "<oldname> <newname>",      do_mv},
{"rm",	    1, 1, 1, "<file>",                   do_rm},
{"mkdir",   1, 1, 1, "<dir>",                    do_mkdir},
{"rmdir",   1, 1, 1, "<dir>",                    do_rmdir},
{"stat",    1, 1, 1, "<file>",                   do_stat},
{"statfs",  1, 0, 0, "",                   	 do_statfs},
{"df",      1, 0, 0, "",                   	 do_statfs},
{"timeout", 0, 1, 1, "<seconds>",                do_timeout},
{"debug",   0, 0, 1, "[subsystem]",              do_debug},
{"whoami",  1, 0, 0, "",                         do_whoami},
{"exec",    1, 4, 100, "<infile> <outfile> <errfile> <command> [args]",do_exec},
{"run",     1, 1, 100, "<command> [args] [<infile] [>outfile] [2>errfile]",do_run},
{"md5",     1, 1, 1,  "<path>",                  do_md5},
{"help",    0, 0, 0, "",                         do_help},
{"exit",    0, 0, 0, "",                         do_quit},
{"quit",    0, 0, 0, "",                         do_quit},
{"thirdput",1, 3, 3, "<file> <3rdhost> <3rdfile>",do_thirdput},
{"lsalloc", 1, 0, 1, "[path]",                   do_lsalloc},
{"mkalloc", 1, 2, 2, "<path> <size>",            do_mkalloc},

{"whoareyou",1, 1, 1, "<hostname>", do_whoareyou},
{"creategroup", 1, 1, 1,  "<groupname>", do_creategroup},
{"deletegroup", 1, 1, 1,  "<groupname>", do_rm},
{"removegroup", 1, 1, 1,  "<groupname>", do_rm},
{"adduser", 1, 2, 2,  "<groupname> <username>", do_adduser},
{"removeuser", 1, 2, 2,  "<groupname> <username>", do_removeuser},
{"membership", 1, 1, 1,  "<groupname>", do_membership},
{"setpolicy", 1, 1, 5, "-f <file cache duration> -d <decision cache duration> <groupname>", do_set_policy},
{0,0,0,0,0},
};

static void show_version( const char *cmd )
{
	printf("%s version %d.%d.%d built by %s@%s on %s at %s\n",cmd,CCTOOLS_VERSION_MAJOR,CCTOOLS_VERSION_MINOR,CCTOOLS_VERSION_MICRO,BUILD_USER,BUILD_HOST,__DATE__,__TIME__);
}

static void show_help( const char *cmd )
{
	printf("use: %s [options] [hostname] [command]\n",cmd);
	printf("where options are:\n");
	printf(" -a <flag>  Require this authentication mode.\n");
	printf(" -d <flag>  Enable debugging for this subsystem.\n");
	printf(" -v         Show program version.\n");
	printf(" -h         This message.\n");
	printf(" -l         Long transfer information.\n");
}
 

int main( int argc, char *argv[] )
{	
#ifdef HAS_LIBREADLINE
	char *temp;
#endif
	int did_explicit_auth = 0;
	int interactive_mode;
	char prompt[CHIRP_LINE_MAX];
	char line[CHIRP_LINE_MAX];
	char **user_argv=0;
	int user_argc;
	char c;
	int result;

	debug_config(argv[0]);

	while((c=getopt(argc,argv,"+a:d:vhl"))!=(char)-1) {
		switch(c) {
			case 'a':
				auth_register_byname(optarg);
				did_explicit_auth = 1;
				break;
			case 'd':
				debug_flags_set(optarg);
				break;
			case 'v':
				show_version(argv[0]);
				exit(0);
				break;
			case 'h':
				show_help(argv[0]);
				exit(0);
				break;
		        case 'l':
			        long_information = 1;
				break;

		}
	}
 
	if(!did_explicit_auth) auth_register_all();
	interactive_mode = isatty(0);
	getcwd(current_local_dir,CHIRP_PATH_MAX);

	if(optind<argc) {
		stoptime = time(0) + timeout;
		if(do_open(1,&argv[optind-1])) {
			fprintf(stderr,"couldn't open %s: %s\n",argv[optind],strerror(errno));
			return 1;
		}
	}

	if((argc-optind)>1) {
		return !process_command(argc-optind-1,&argv[optind+1]);
	}

	while(1) {
		if(interactive_mode) {
			sprintf(prompt," chirp:%s:%s> ",current_host,current_remote_dir);
		} else {
			prompt[0] = 0;
		}

#ifdef HAS_LIBREADLINE
		temp = readline(prompt);		
		if(!temp) break;
		strcpy(line,temp);
		free(temp);
#else
		printf("%s",prompt);
		fflush(stdout);
		if(!fgets(line,CHIRP_LINE_MAX,stdin)) break;
#endif

		if(!line[0]) continue;
#ifdef HAS_LIBREADLINE
		add_history(line);
#endif

		if(user_argv) free(user_argv);
		string_split(line,&user_argc,&user_argv);
		if(user_argc==0) continue;

		result = process_command(user_argc,user_argv);
		if(!interactive_mode && !result) break;
	}

	if(result) {
		return 0;
	} else {
		return 1;
	}
}

static int process_command( int argc, char **argv )
{
	int i;

	for(i=0;list[i].name;i++) {
		if(!strcmp(list[i].name,argv[0])) {
			if(!client && list[i].must_be_open) {
				printf("not connected\n");
				return 0;
			} else if( ((argc-1)>=list[i].minargs) && ((argc-1)<=list[i].maxargs ) ) {
				stoptime = time(0)+timeout;
				if(list[i].handler(argc,argv)<0) {
					fprintf(stderr,"couldn't %s: %s\n",argv[0],strerror(errno));
					return 0;
				} else {
					return 1;
				}
			} else {
				printf("use: %s %s\n",argv[0],list[i].help);
				return 0;
			}
		}
	}

	printf("unknown command: %s  (try 'help')\n",argv[0]);
	return 0;
}

static int do_open( int argc, char **argv )
{
	do_close(0,0);
	client = chirp_client_connect(argv[1],1,stoptime);
	if(!client) {
		return -1;
	} else {	  
		strcpy(current_host,argv[1]);
		strcpy(current_remote_dir,"/");
		printf("connected to %s as ",current_host);
		do_whoami(0,0);
		return 0;
	}
}

static int do_close( int argc, char **argv )
{
	if(client) {
		chirp_client_disconnect(client);
		current_host[0] = 0;
		current_remote_dir[0] = 0;
		client = 0;
	}
	return 0;
}

static void complete_local_path( const char *file, char *full_path )
{
	char temp[CHIRP_PATH_MAX];
	if(file[0]!='/') {
		sprintf(temp,"%s/%s",current_local_dir,file);
	} else {
		strcpy(temp,file);
	}
	string_collapse_path(temp,full_path,1);
}

static void complete_remote_path( const char *file, char *full_path )
{
	char temp[CHIRP_PATH_MAX];
	if(file[0]!='/') {
		sprintf(temp,"%s/%s",current_remote_dir,file);
	} else {
		strcpy(temp,file);
	}
	string_collapse_path(temp,full_path,1);
}

static int do_cat( int argc, char **argv )
{
	INT64_T actual;
	int i;

	for(i=1;i<argc;i++) {
		char full_path[CHIRP_PATH_MAX];
		complete_remote_path(argv[i],full_path);
		actual = chirp_client_getfile(client,full_path,stdout,stoptime);
		if(actual<0) {
			fprintf(stderr,"%s: %s\n",full_path,strerror(errno));
		}
	}

	return 0;
}

static int do_cd( int argc, char **argv )
{
	char full_path[CHIRP_PATH_MAX];
	struct chirp_stat info;
	complete_remote_path(argv[1],full_path);
	if(chirp_client_stat(client,full_path,&info,stoptime)<0) {
		return -1;
	} else {
		if(S_ISDIR(info.cst_mode)) {
			string_collapse_path(full_path,current_remote_dir,1);
			return 0;
		} else {
			errno = ENOTDIR;
			return -1;
		}
	}
}

static int do_lcd( int argc, char **argv )
{
	char full_path[CHIRP_PATH_MAX];
	complete_local_path(argv[1],full_path);
	if(chdir(full_path)==0) {
		getcwd(current_local_dir,CHIRP_PATH_MAX);
		return 0;
	} else {
		return -1;
	}
}

static int do_pwd( int argc, char **argv )
{
	printf("%s\n",current_remote_dir);
	return 0;
}

static int do_lpwd( int argc, char **argv )
{
	printf("%s\n",current_local_dir);
	return 0;
}

static int do_get_one_dir( const char *source_file, const char *target_file, int mode )
{
	char new_source_file[CHIRP_PATH_MAX];
	char new_target_file[CHIRP_PATH_MAX];
	struct list *work_list;
	const char *name;
	int result;

	work_list = list_create();

	result = mkdir(target_file,mode);
	if(result==0 || errno==EEXIST ) {
		result = chirp_client_opendir(client,source_file,stoptime);
		if(result==0) {
			while((name=chirp_client_readdir(client,stoptime))) {
				if(!strcmp(name,".")) continue;
				if(!strcmp(name,"..")) continue;
				list_push_tail(work_list,xstrdup(name));
			}
			while((name=list_pop_head(work_list))) {
				sprintf(new_source_file,"%s/%s",source_file,name);
				sprintf(new_target_file,"%s/%s",target_file,name);
				result = do_get_recursive(new_source_file,new_target_file);
				free((char*)name);
				if(result<0) break;
			}
		} else {
			result = -1;
		}
	} else {
		result = -1;
	}

	while((name=list_pop_head(work_list))) free((char*)name);
	list_delete(work_list);

	return result;
}

static int do_get_one_link( const char *source_file, const char *target_file )
{
	char linkdata[CHIRP_PATH_MAX];
	int result;

	result = chirp_client_readlink(client,source_file,linkdata,sizeof(linkdata),stoptime);
	if(result>0) {
		linkdata[result] = 0;
		unlink(target_file);
		result = symlink(linkdata,target_file);
		if(result>=0) result = 0;
	}

	return result;
}

static int do_get_one_file( const char *source_file, const char *target_file, int mode, INT64_T length )
{
	FILE *file;
	int save_errno;
	INT64_T actual;

	file = fopen64(target_file,"w");
	if(!file) return -1;

	fchmod(fileno(file),mode);

	actual = chirp_client_getfile(client,source_file,file,stoptime);
	if(actual!=length) {
		save_errno = errno;
		fclose(file);
		errno = save_errno;
		return -1;
	}

	if(length>=0) {
		fclose(file);
		return 0;
	} else {
		save_errno = errno;
		fclose(file);
		errno = save_errno;
		return -1;
	}
}

static int do_get_recursive( const char *source_file, const char *target_file )
{
	int result;
	struct chirp_stat info;

	result = chirp_client_lstat(client,source_file,&info,stoptime);
	if(result<0) {
		result = -1;
	} else {
		if(S_ISLNK(info.cst_mode)) {
			printf("link %s\n->   %s\n",source_file,target_file);
			result = do_get_one_link(source_file,target_file);
		} else if(S_ISDIR(info.cst_mode)) {
			printf("dir  %s\n->   %s\n",source_file,target_file);
			result = do_get_one_dir(source_file,target_file,info.cst_mode);
		} else if(S_ISREG(info.cst_mode)) {
			printf("file %s [%sB]\n->   %s\n",source_file,string_metric(info.cst_size,-1,0),target_file);
			result = do_get_one_file(source_file,target_file,info.cst_mode,info.cst_size);
			xfer_bytes += info.cst_size;
		} else {
			printf("???  %s\n->   %s\n",source_file,target_file);
			result = 0;
		}
		xfer_items++;
	}

	if(result<0) {
		printf("couldn't get %s: %s\n",source_file,strerror(errno));
	}

	return result;
}

static int do_get( int argc, char **argv )
{
	char target_full_path[CHIRP_PATH_MAX];
	char source_full_path[CHIRP_PATH_MAX];
	timestamp_t start, stop;
	double elapsed;
	int result;

	if(!argv[2]) argv[2]=(char*)string_basename(argv[1]);

	complete_remote_path(argv[1],source_full_path);
	complete_local_path(argv[2],target_full_path);

	xfer_items = 0;
	xfer_bytes = 0;

	start = timestamp_get();
	result = do_get_recursive(source_full_path,target_full_path);
	stop = timestamp_get();

	elapsed = (stop-start)/1000000.0;

	printf("%lld files and %sB read in %.2fs ",
		xfer_items,
		string_metric(xfer_bytes,-1,0),
		elapsed);

	printf(" (%sB/s)\n",string_metric(xfer_bytes/elapsed,-1,0));

	return result;
}


static int do_put_one_dir( const char *source_file, const char *target_file, int mode )
{
	char new_source_file[CHIRP_PATH_MAX];
	char new_target_file[CHIRP_PATH_MAX];
	struct list *work_list;
	const char *name;
	int result;
	struct dirent *d;
	DIR *dir;

	work_list = list_create();

	result = chirp_client_mkdir(client,target_file,mode,stoptime);
	if(result==0 || errno==EEXIST ) {
		result = 0;
		dir = opendir(source_file);
		if(dir) {
			while((d=readdir(dir))) {
				if(!strcmp(d->d_name,".")) continue;
				if(!strcmp(d->d_name,"..")) continue;
				list_push_tail(work_list,xstrdup(d->d_name));
			}
			closedir(dir);
			while((name=list_pop_head(work_list))) {
				sprintf(new_source_file,"%s/%s",source_file,name);
				sprintf(new_target_file,"%s/%s",target_file,name);
				result = do_put_recursive(new_source_file,new_target_file);
				free((char*)name);
				if(result<0) break;
			}
		} else {
			result = -1;
		}
	} else {
		result = -1;
	}

	while((name=list_pop_head(work_list))) free((char*)name);
	list_delete(work_list);

	return result;
}

static int do_put_one_link( const char *source_file, const char *target_file )
{
	char linkdata[CHIRP_PATH_MAX];
	int result;

	result = readlink(source_file,linkdata,sizeof(linkdata));
	if(result>0) {
		char linkdatafull[CHIRP_PATH_MAX];
		linkdata[result] = 0;
		complete_local_path(linkdata,linkdatafull);
		chirp_client_unlink(client,target_file,stoptime);
       		result = chirp_client_symlink(client,linkdatafull,target_file,stoptime);
		if(result>=0) result = 0;
	}

	return result;
}

static int do_put_one_file( const char *source_file, const char *target_file, int mode, INT64_T length )
{
	FILE *file;
	int save_errno;

	file = fopen64(source_file,"r");
	if(!file) return -1;

	length = chirp_client_putfile(client,target_file,file,mode,length,stoptime);
	if(length<0) {
		save_errno = errno;
		fclose(file);
		errno = save_errno;
		return -1;
	}

	if(length>=0) {
		fclose(file);
		//return 0;
		return length;
	} else {
		save_errno = errno;
		fclose(file);
		errno = save_errno;
		return -1;
	}
}

static int do_put_recursive( const char *source_file, const char *target_file )
{
	int result;
	struct stat64 info;

	result = lstat64(source_file,&info);
	if(result<0) {
		result = -1;
	} else {
		if(S_ISLNK(info.st_mode)) {
			printf("link %s\n->   %s\n",source_file,target_file);
			result = do_put_one_link(source_file,target_file);
		} else if(S_ISDIR(info.st_mode)) {
			printf("dir  %s\n->   %s\n",source_file,target_file);
			result = do_put_one_dir(source_file,target_file,0700);
		} else if(S_ISREG(info.st_mode)) {
			printf("file %s [%sB]\n->   %s\n",source_file,string_metric(info.st_size,-1,0),target_file);
			result = do_put_one_file(source_file,target_file,info.st_mode,info.st_size);
			//xfer_bytes += info.st_size;
			if( result >= 0 )
			  xfer_bytes += result;
		} else {
			printf("???  %s\n->   %s\n",source_file,target_file);
			result = 0;
		}
		if( result >= 0 )
		  xfer_items++;
	}

	if(result<0) {
		printf("couldn't put %s: %s\n",source_file,strerror(errno));
	}

	return result;
}

static int do_put( int argc, char **argv )
{
	char target_full_path[CHIRP_PATH_MAX];
	char source_full_path[CHIRP_PATH_MAX];
	timestamp_t start, stop;
	double elapsed;
	int result;

	if(!argv[2]) argv[2]=(char*)string_basename(argv[1]);

	complete_local_path(argv[1],source_full_path);
	complete_remote_path(argv[2],target_full_path);

	xfer_items = 0;
	xfer_bytes = 0;

	start = timestamp_get();
	result = do_put_recursive(source_full_path,target_full_path);
	stop = timestamp_get();

	elapsed = (stop-start)/1000000.0;

	printf("%lld files and %sB written in %.2fs ",
	       xfer_items,
	       string_metric(xfer_bytes,-1,0),
	       elapsed);
	printf(" (%sB/s)\n",string_metric(xfer_bytes/elapsed,-1,0));

	return result;
}

static int do_setacl( int argc, char **argv )
{
	char full_path[CHIRP_PATH_MAX];
	complete_remote_path(argv[1],full_path);

	if(!strcmp(argv[3],"read")) argv[3] = "rl";
	if(!strcmp(argv[3],"write")) argv[3] = "rwl";
	if(!strcmp(argv[3],"admin")) argv[3] = "rwlva";
	if(!strcmp(argv[3],"reserve")) argv[3] = "lv";
	if(!strcmp(argv[3],"none")) argv[3] = ".";

	return chirp_client_setacl(client,full_path,argv[2],argv[3],stoptime);
}

static int do_getacl( int argc, char **argv )
{
	char full_path[CHIRP_PATH_MAX];
	const char *name;
	int result;

	if(!argv[1]) argv[1] = ".";
	complete_remote_path(argv[1],full_path);

	result = chirp_client_openacl(client,full_path,stoptime);
	if(result==0) {
		while((name=chirp_client_readacl(client,stoptime))) {
			printf("%s\n",name);
		}
	}

	return result;
}

static int print_long_list( const char *shortname, const char *longname )
{
	char linkname[CHIRP_PATH_MAX];
	struct chirp_stat info;
	int result;
	const char *type;

	result = chirp_client_lstat(client,longname,&info,stoptime);
	if(result<0) {
		printf("couldn't lstat %s: %s\n",shortname,strerror(errno));
		return -1;
	}

	linkname[0] = 0;

	if(S_ISLNK(info.cst_mode)) {
		result = chirp_client_readlink(client,longname,linkname,sizeof(linkname),stoptime);
		if(result>0) {
			linkname[result] = 0;
			strcat(linkname,"\n");
		} else {
			strcpy(linkname,"???");
		}
		type = "link";
	} else if(S_ISDIR(info.cst_mode)) {
		type = "dir ";
	} else if(S_ISREG(info.cst_mode)) {
		type = "file";
	} else {
		type = "????";
	}

	time_t t = info.cst_ctime;
	printf("%4s %8lld %-40s %24s",type,info.cst_size,shortname,linkname[0] ? linkname : ctime(&t));

	return 0;
}

static int do_ls( int argc, char **argv )
{
	char full_path[CHIRP_PATH_MAX];
	struct list *work_list;
	const char *name;
	int long_mode = 0;
	int all_mode = 0;
	int result;

	const char *options = argv[1];
	const char *file = argv[2];

	if(!options) options="";
	if(!file) file="";

	if(options[0]!='-') {
		file = options;
		options = "-";
	}

	options++;
	while(*options) {
		switch(*options) {
			case 'l':
				long_mode = 1;
				break;
			case 'a':
				all_mode = 1;
				break;
			default:
				printf("ls: unknown option: %c\n",*options);
				errno = EINVAL;
				return -1;
		}
		options++;
	}

	if(!file[0]) file = ".";
	complete_remote_path(file,full_path);

	work_list = list_create();

	result = chirp_client_opendir(client,full_path,stoptime);
	if(result==0) {
		while((name=chirp_client_readdir(client,stoptime))) {
			list_push_tail(work_list,xstrdup(name));
		}
		while((name=list_pop_head(work_list))) {
			char new_full_path[CHIRP_PATH_MAX];
			sprintf(new_full_path,"%s/%s",full_path,name);

			if(!all_mode && name[0]=='.') {
				/* do nothing */
			} else {
				if(long_mode) {
					print_long_list(name,new_full_path);
				} else {
					printf("%s\n",name);
				}
			}
			free((char*)name);
		}
	}

	while((name=list_pop_head(work_list))) free((char*)name);
	list_delete(work_list);

	return result;
}

static int do_rm_recursive( const char *file )
{
	struct list *work_list;
	const char *name;
	int result;

	work_list = list_create();

	result = chirp_client_unlink(client,file,stoptime);
	if(result==0 || errno==ENOENT) {
		result = 0;
	} else if(result<0 && errno==EISDIR) {
		result = chirp_client_opendir(client,file,stoptime);
		if(result==0) {
			while((name=chirp_client_readdir(client,stoptime))) {
				if(!strcmp(name,".")) continue;
				if(!strcmp(name,"..")) continue;
				list_push_tail(work_list,xstrdup(name));
			}
			while((name = list_pop_head(work_list))) {
				char newfile[CHIRP_PATH_MAX];
				sprintf(newfile,"%s/%s",file,name);
				result = do_rm_recursive(newfile);
				free((char*)name);
				if(result<0) break;
			}
			result = chirp_client_rmdir(client,file,stoptime);
		} else {
			result = -1;
		}
	}

	if(result==0) {
		printf("removed: %s\n",file);
	} else {
		printf("couldn't rm %s: %s\n",file,strerror(errno));
	}

	while((name=list_pop_head(work_list))) free((char*)name);
	list_delete(work_list);

	return result;
}

static int do_rm( int argc, char **argv )
{
	char full_path[CHIRP_PATH_MAX];
	complete_remote_path(argv[1],full_path);
	return do_rm_recursive(full_path);
}

static int do_rmdir( int argc, char **argv )
{
	char full_path[CHIRP_PATH_MAX];
	complete_remote_path(argv[1],full_path);
	return chirp_client_rmdir(client,full_path,stoptime);
}

static int do_mkdir( int argc, char **argv )
{
	char full_path[CHIRP_PATH_MAX];
	int result;

	complete_remote_path(argv[1],full_path);
	result = chirp_client_mkdir(client,full_path,0777,stoptime);

	if(result<0 && errno==EEXIST) result = 0;

	return result;
}

static int do_stat( int argc, char **argv )
{
	char full_path[CHIRP_PATH_MAX];
	struct chirp_stat info;
	time_t t;

	complete_remote_path(argv[1],full_path);

	if(chirp_client_stat(client,full_path,&info,stoptime)<0) {
		return -1;
	} else {
		printf("device:  %lld\n",info.cst_dev);
		printf("inode:   %lld\n",info.cst_ino);
		printf("mode:    %04llo\n",info.cst_mode);
		printf("nlink:   %lld\n",info.cst_nlink);
		printf("uid:     %lld\n",info.cst_uid);
		printf("gid:     %lld\n",info.cst_gid);
		printf("rdevice: %lld\n",info.cst_rdev);
		printf("size:    %lld\n",info.cst_size);
		printf("blksize: %lld\n",info.cst_blksize);
		printf("blocks:  %lld\n",info.cst_blocks);
		t = info.cst_atime;
		printf("atime:   %s",ctime(&t));
		t = info.cst_mtime;
		printf("mtime:   %s",ctime(&t));
		t = info.cst_ctime;
		printf("ctime:   %s",ctime(&t));
		return 0;
	}
 }

static int do_statfs( int argc, char **argv )
{
	struct chirp_statfs info;

	if(chirp_client_statfs(client,"/",&info,stoptime)<0) {
		return -1;
	} else {
		printf("/\n");
		printf("%sB TOTAL\n",string_metric(info.f_blocks*info.f_bsize,-1,0));
		printf("%sB INUSE\n",string_metric((info.f_blocks-info.f_bfree)*info.f_bsize,-1,0));
		return 0;
	}
}

static int do_mv( int argc, char **argv )
{
	char old_full_path[CHIRP_PATH_MAX];
	char new_full_path[CHIRP_PATH_MAX];
	complete_remote_path(argv[1],old_full_path);
	complete_remote_path(argv[2],new_full_path);
	return chirp_client_rename(client,old_full_path,new_full_path,stoptime);
}

static int do_debug( int argc, char **argv )
{
	if(argv[1]) {
		if(debug_flags_set(argv[1])) {
			return 0;
		} else {
			printf("Valid debugging flags are: ");
			debug_flags_print(stdout);
			printf("\n");
			errno = EINVAL;
			return -1;
		}
	} else {
		debug_flags_clear();
		return 0;
	}	
	
}

static int do_whoami( int argc, char **argv )
{
	char name[CHIRP_LINE_MAX];
	int result;

	result = chirp_client_whoami(client,name,sizeof(name),stoptime);
	if(result>=0) {
		name[result] = 0;
		printf("%s\n",name);
	}
	return result;
}

static int do_whoareyou( int argc, char **argv )
{
	char name[CHIRP_LINE_MAX];
	int result;

	result = chirp_client_whoareyou(client,argv[1],name,sizeof(name),
					stoptime);
	if(result>=0) {
		name[result] = 0;
		printf("%s\n",name);
	}
	return result;
}

static int do_exec( int argc, char **argv )
{
	char infile[CHIRP_LINE_MAX];
	char outfile[CHIRP_LINE_MAX];
	char errfile[CHIRP_LINE_MAX];
	char cmd[CHIRP_LINE_MAX];
	int i;

	if(!strcmp(argv[1],"-")) {
		strcpy(infile,argv[1]);
	} else {
		complete_remote_path(argv[1],infile);
	}

	if(!strcmp(argv[2],"-")) {
		strcpy(outfile,argv[2]);
	} else {
		complete_remote_path(argv[2],outfile);
	}

	if(!strcmp(argv[3],"-")) {
		strcpy(errfile,argv[3]);
	} else {
		complete_remote_path(argv[3],errfile);
	}

	complete_remote_path(argv[4],cmd);

	for(i=5;i<argc;i++) {
		strcat(cmd," ");
		strcat(cmd,argv[i]);
	}

	return chirp_client_exec(client,infile,outfile,errfile,cmd,stoptime);
}

static int do_run( int argc, char **argv )
{
	char cmd[CHIRP_LINE_MAX];
	char infile[CHIRP_LINE_MAX];
	char outfile[CHIRP_LINE_MAX];
	char errfile[CHIRP_LINE_MAX];
	const char *inarg=0;
	const char *outarg=0;
	const char *errarg=0;
	int result;
	int i;

	cmd[0] = 0;

	for(i=1;i<argc;i++) {
		if(argv[i][0]=='<') {
			inarg = &argv[i][1];
		} else if(argv[i][0]=='>') {
			outarg = &argv[i][1];
		} else if(!strncmp(argv[i],"1>",2)) {
			outarg = &argv[i][2];
			if(!strcmp(outarg,"&2")) outarg = errarg;
		} else if(!strncmp(argv[i],"2>",2)) {
			errarg = &argv[i][2];		
			if(!strcmp(errarg,"&1")) errarg = outarg;
		} else {
			if(!cmd[0]) {
				complete_remote_path(argv[i],cmd);
			} else {
				strcat(cmd," ");
				strcat(cmd,argv[i]);
			}
		}
	}
	

	if(inarg) {
		complete_remote_path(inarg,infile);
	} else {
		strcpy(infile,"-");
	}

	if(outarg) {
		complete_remote_path(outarg,outfile);
	} else {
		complete_remote_path("stdout.txt",outfile);
	}

	if(errarg) {
		complete_remote_path(errarg,errfile);
	} else {
		complete_remote_path("stderr.txt",errfile);
	}

	result = chirp_client_exec(client,infile,outfile,errfile,cmd,stoptime);

	if(result>=256) {
		printf("*** %s exited abnormally with signal %d\n",argv[1],result-256);
	} else if(result>=0) {
		printf("*** %s exited with result %d\n",argv[1],result);
	} else {
		printf("*** exec %s failed: %s\n",argv[1],strerror(errno));
	}

	if(result>=0) {
		if(!outarg) chirp_client_getfile(client,outfile,stdout,stoptime);
		if(!errarg) chirp_client_getfile(client,errfile,stdout,stoptime);
	}

	return 0;
}

static int do_md5( int argc, char **argv )
{
	unsigned char digest[16];
	char full_path[CHIRP_LINE_MAX];
	int result;

	complete_remote_path(argv[1],full_path);

	result = chirp_client_md5(client,full_path,digest,stoptime);
	if(result>0) printf("%s %s\n",md5_string(digest),full_path);
	return result;
}

static int do_timeout( int argc, char **argv )
{
	timeout = atoi(argv[1]);
	printf("Timeout is %d seconds.\n",timeout);
	return 0;
}

static int do_quit( int argc, char **argv )
{
	exit(0);
	return -1;
}

static int do_help( int argc, char **argv )
{
	int i;
	printf("Commands:\n");
	for(i=0;list[i].name;i++) {
		printf("  %8s %s\n",list[i].name,list[i].help);
	}
	printf("\nDebugging subsystems are:\n");
	debug_flags_print(stdout);
	printf("\n");
	return 0;
}

static int do_thirdput( int argc, char **argv ) {
  const char *temp, *path3pfile;
  char full_path[CHIRP_PATH_MAX], path3pcontainer[CHIRP_PATH_MAX];
  char name[CHIRP_LINE_MAX], acl[CHIRP_LINE_MAX], backpath[CHIRP_LINE_MAX];
  char hostsubject[CHIRP_LINE_MAX];
  char *thirdparty = argv[2], *path3p = argv[3];
  struct chirp_client *client3p = 0;
  int result = 0, needtosetacl = 1, tmperr = 0, flags = 0, tmpresult = 0;

  /// PREPARATORY WORK

  if(!argv[1]) argv[1] = ".";
  complete_remote_path(argv[1],full_path);

  sprintf(backpath,"%s/..",path3p);
  string_collapse_path(backpath,path3pcontainer,1);

  //have host issue remote whoami
  result = chirp_client_whoareyou( client, thirdparty, hostsubject,
				   CHIRP_LINE_MAX, stoptime );

  if( result > 0 ) {
    //add null zero to end of string
    if( result < sizeof( hostsubject ) )
      hostsubject[result] = '\0';
    else
      hostsubject[sizeof( hostsubject ) - 1] = '\0';

    //separate path from file
    path3pfile = string_basename( argv[3] );
    
    ////REAL WORK 
    
    //connect to third party
    client3p = chirp_client_connect( thirdparty, 1, stoptime );
  
    if( client3p ) {
      //check for write privilege on third party containing dir
      result = chirp_client_openacl( client3p, path3pcontainer, stoptime );
      if( result >= 0 ) {
	//parse acl for the hostname
	while(( temp = chirp_client_readacl(client3p,stoptime) )) {
	  sscanf(temp, "%s %s\n", name, acl);
	  if( strcmp( name, hostsubject ) == 0 ) {
	    //convert the ACLs to flags
	    flags = chirp_acl_text_to_flags ( acl );
	    if( flags & CHIRP_ACL_WRITE )
	      needtosetacl = 0;
	    else
	      needtosetacl = -1;
	    break;
	  }
	}
	
	if( needtosetacl ) {
	  //if the host was not in the ACL at all, give a default empty ACL
	  if( needtosetacl == 1 )
	    flags = 0;
	  
	  result = chirp_client_setacl( client3p, path3pcontainer, hostsubject,
					chirp_acl_flags_to_text
					( flags | CHIRP_ACL_WRITE ),
					stoptime );
	  if( result < 0 ) {
	    //save error number
	    tmperr = errno;
	    chirp_client_disconnect( client3p );
	    errno = tmperr;
	    return result;
	  }
	}
	
	//issue thirdput RPC to second party
	result = chirp_client_thirdput( client, full_path, thirdparty, path3p,
					stoptime );
	
	//save errno if an error occured
	if( result < 0)
	  fprintf( stderr, "couldn't thirdput: %s\n", strerror(errno) );
	
	if( needtosetacl ) {
	  //remove temporary write privilege on path to second party
	  if( flags == 0 )
	    tmpresult = chirp_client_setacl( client3p, path3pcontainer,
					     hostsubject, ".", stoptime );
	  else
	    tmpresult = chirp_client_setacl( client3p, path3pcontainer,
					     hostsubject,
					     chirp_acl_flags_to_text( flags ),
					     stoptime );
	  
	  if( tmpresult < 0 )
	    tmperr = errno;
	}
      }
      else {
	tmperr = errno;
      }
      
      //close third party connection
      chirp_client_disconnect( client3p );
      if( tmperr ){
	errno = tmperr;
	return -1;
      }
      else
	return 0;
    }
    else {
      return -1;
    }
  }
  else
    return -1;
}

static int do_mkalloc( int argc, char **argv )
{
	char full_path[CHIRP_PATH_MAX];
	complete_remote_path(argv[1],full_path);	
	return chirp_client_mkalloc(client,full_path,string_metric_parse(argv[2]),0700,stoptime);
}

static int do_lsalloc( int argc, char **argv )
{
	char full_path[CHIRP_PATH_MAX];
	char alloc_path[CHIRP_PATH_MAX];
	INT64_T total, inuse;
	int result;

	if(argc!=2) argv[1] = ".";

	complete_remote_path(argv[1],full_path);	

	result = chirp_client_lsalloc(client,full_path,alloc_path,&total,&inuse,stoptime);

	if(result>=0) {
		printf("%s\n",alloc_path);
		printf("%sB TOTAL\n",string_metric(total,-1,0));
		printf("%sB INUSE\n",string_metric(inuse,-1,0));
	}

	return result;
}

static int do_creategroup(int argc, char** argv)
{
  char full_path[CHIRP_PATH_MAX];

  if (argc < 2)
    return -1;

  complete_remote_path(argv[1], full_path);

  return chirp_client_creategroup(client, full_path, stoptime);
}

static int do_adduser(int argc, char** argv)
{
  char full_path[CHIRP_PATH_MAX];

  if (argc < 3)
    return -1;

  complete_remote_path(argv[1], full_path);
     
  return chirp_client_adduser(client, full_path, argv[2], stoptime);
}

static int do_removeuser(int argc, char** argv)
{
  char full_path[CHIRP_PATH_MAX];

  if (argc < 3)
    return -1;

  complete_remote_path(argv[1], full_path);
     
  return chirp_client_removeuser(client, full_path, argv[2], stoptime);
}

static int do_membership(int argc, char** argv)
{
  static char* name;
  static char full_path[CHIRP_PATH_MAX];
  int result;

  complete_remote_path(argv[1], full_path);

  result = chirp_client_open_group(client, full_path, stoptime);
  if (result == 0)
    {
      while((name = chirp_client_get_group_members(client, stoptime)) != NULL) {
	printf("%s\n", name);
      }
    }
  return 1;
}

static int do_set_policy(int argc, char** argv)
{
  static char full_path[CHIRP_PATH_MAX];
  int result;
  long int file_policy = 0;
  long int dec_policy = 0;
  char c;

  optind = opterr = 0;
  while((c=getopt(argc,argv,"+d:f:D:F")) != (char)-1) {
          switch(c) {
                  case 'd':
                          dec_policy = atol(optarg);
                          break;
                  case 'D':
                          dec_policy = atol(optarg);
                          break;
                  case 'f':
                          file_policy = atol(optarg);
                          break;
                  case 'F':
                          file_policy = atol(optarg);
                          break;
    }
  }

  if (argv[optind] == NULL)  /* no group name specified */
    {
      errno = EINVAL;
      return -1;
    }
  complete_remote_path(argv[optind], full_path);

  result = chirp_client_set_policy(client, full_path, file_policy, dec_policy, stoptime);

  return result;
}

