/*
Copyright (C) 2003-2004 Douglas Thain and the University of Wisconsin
Copyright (C) 2005- The University of Notre Dame
This software is distributed under a BSD-style license.
See the file COPYING for details.
*/

#include "file_cache.h"
#include "create_dir.h"
#include "hash_table.h"
#include "debug.h"

#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <dirent.h>
#include <limits.h>

/* Cygwin does not have 64-bit I/O, while Darwin has it by default. */
 
#if CCTOOLS_OPSYS_CYGWIN || CCTOOLS_OPSYS_DARWIN
#define stat64 stat
#define open64 open
#define mkstemp64 mkstemp
#endif

struct file_cache {
	char *root;
};

static void cached_name( struct file_cache *c, const char *path, char *lpath )
{
	unsigned val = hash_string(path);
	sprintf(lpath,"%s/%02x/%08x",c->root,(val>>24)&0xff,val);
}

static void txn_name( struct file_cache *c, const char *path, char *cpath )
{
	unsigned val = hash_string(path);
	sprintf(cpath,"%s/txn/%08x.XXXXXX",c->root,val);
}

struct file_cache *file_cache_init( const char *root )
{
	char path[PATH_MAX];
	struct stat64 buf;
	int result, i;

	struct file_cache *f = malloc(sizeof(*f));
	if(!f) return 0;

	f->root = strdup(root);
	if(!f->root) {
		free(f);
		return 0;
	}

	sprintf(path,"%s/ff",root);
	result = stat64(path,&buf);
	if(result!=0) {
		if(!create_dir(path,0777)) goto failure;
		sprintf(path,"%s/txn",root);
		if(!create_dir(path,0777)) goto failure;
		for(i=0;i<=0xff;i++) {
			sprintf(path,"%s/%02x",root,i);
			if(!create_dir(path,0777)!=0) goto failure;
		}
	}

	return f;

	failure:
	file_cache_fini(f);
	return 0;
}

void file_cache_fini( struct file_cache *f )
{
	if(f) {
		free(f->root);
		free(f);
	}
}

void file_cache_cleanup( struct file_cache *f )
{
	char path[PATH_MAX];
	struct dirent *d;
	DIR *dir;

	sprintf(path,"%s/txn",f->root);

	dir = opendir(path);
	if(!dir) return;

	while((d=readdir(dir))) {
		if(!strcmp(d->d_name,".")) continue;
		if(!strcmp(d->d_name,"..")) continue;
		sprintf(path,"%s/txn/%s",f->root,d->d_name);
		unlink(path);
	}

	closedir(dir);
}

int file_cache_contains( struct file_cache *c, const char *path, char *lpath )
{
	struct stat64 buf;
	cached_name(c,path,lpath);
	return stat64(lpath,&buf);
}

int file_cache_open( struct file_cache *c, const char *path, char *lpath, INT64_T size, time_t mtime )
{
	struct stat64 buf;
	cached_name(c,path,lpath);
	if(stat64(lpath,&buf)==0) {
		if( (size==0 || (size==buf.st_size)) && ( (mtime==0) || (buf.st_mtime>=mtime) )) {
			debug(D_CACHE,"hit %s",path);
			return open64(lpath,O_RDWR,0);
		} else {
			debug(D_CACHE,"stale %s",path);
			errno = ENOENT;
			return -1;
		}
	} else {
		debug(D_CACHE,"miss %s",path);
		return -1;
	}
}

int  file_cache_delete( struct file_cache *f, const char *path )
{
	char lpath[PATH_MAX];
	cached_name(f,path,lpath);
	debug(D_CACHE,"remove %s",path);
	return unlink(lpath);
}

int file_cache_begin( struct file_cache *f, const char *path, char *txn )
{
	int result;
	txn_name(f,path,txn);
	result = mkstemp64(txn);
	if(result>=0) {
		debug(D_CACHE,"add %s",path);
		fchmod(result,0700);
	}
	return result;
}

int file_cache_abort( struct file_cache *f, const char *path, const char *txn )
{
	return unlink(txn);
}

int file_cache_commit( struct file_cache *f, const char *path, const char *txn )
{
	char lpath[PATH_MAX];
	cached_name(f,path,lpath);
	return rename(txn,lpath);
}
